﻿using System;
using System.Collections;
using System.Collections.Generic;

namespace HeuristicLab.Algorithms.PushGP.Stack
{
    /// <summary>
    /// While Push's stacks are generally treated as genuine stacks---that is, inclassions take their arguments from the tops of
    /// the stacks and push their results onto the tops of the stacks---a few inclassions (like YANK and SHOVE) do allow direct access
    /// to "deep" stack elements by means of integer indices. To this extent the stacks can be used as general, random access memory
    /// classures. This is one of the features that ensures the Turing-completeness of Push (another being the arbitrary name/value
    /// bindings supported by the NAME data type and DEFINE methods; see below).
    /// </summary>
    /// <typeparam name="T">The item type of the collection.</typeparam>
    public class PushGPStack<T> : IStack<T>
    {
        private readonly List<T> data;
        private const string Delimiter = " ";

        public PushGPStack(int capacity = 0)
        {
            data = new List<T>();
        }

        public T Top { get { return data[Count - 1]; } }
        public T TopOrDefault { get { return Count > 0 ? data[Count - 1] : default(T); } }
        public T Bottom { get { return data[0]; } }
        public T BottomOrDefault { get { return Count > 0 ? data[0] : default(T); } }
        public int Count { get { return data.Count; } }
        public int Capacity { get { return data.Capacity; } }
        public bool IsEmpty { get { return Count == 0; } }

        public bool IsReadOnly { get { return false; } }

        public void Add(T item)
        {
            Push(item);
        }

        public void Clear()
        {
            data.Clear();
        }

        public bool Contains(T item)
        {
            return data.Contains(item);
        }

        public void CopyTo(T[] array, int arrayIndex)
        {
            data.CopyTo(array, arrayIndex);
        }

        public IEnumerator<T> GetEnumerator()
        {
            return data.GetEnumerator();
        }

        public T ElementAt(int index)
        {
            return data[index];
        }


        public T ReverseElementAt(int offset)
        {
            return data[Count - 1 - offset];
        }

        public void SetTop(T value)
        {
            data[Count - 1] = value;
        }

        public T this[int key]
        {
            get { return data[key]; }
            set { data[key] = value; }
        }

        public void Swap(int count)
        {
            var top = Top;
            var bottomIndex = Count - count;

            for (var i = Count - 1; i > bottomIndex; i--)
            {
                data[i] = data[i - 1];
            }

            data[bottomIndex] = top;
        }

        public void Yank(int index)
        {
            var item = ElementAt(index);
            data.RemoveAt(index);
            data.Add(item);
        }

        public T Pop()
        {
            var value = data[Count - 1];
            data.RemoveAt(Count - 1);

            return value;
        }

        public T[] Pop(int count)
        {
            var startIndex = Count - count;

            var items = new T[count];
            data.CopyTo(startIndex, items, 0, count);
            Remove(count);

            return items;
        }

        public void Push(T item)
        {
            data.Add(item);
        }

        public void Push(object[] items)
        {
            for (var i = 0; i < items.Length; i++)
            {
                data.Add((T)items[i]);
            }
        }

        public void PushResult(int count, Func<T[], T> templateFunc)
        {
            var startIndex = Count - count;
            var items = new T[count];

            data.CopyTo(startIndex, items, 0, count);
            Remove(count - 1);

            data[Count - 1] = templateFunc(items);
        }

        public void Push(params T[] items)
        {
            for (var i = 0; i < items.Length; i++)
            {
                data.Add(items[i]);
            }
        }

        public void Push(IEnumerable<T> items)
        {
            data.AddRange(items);
        }

        public void Insert(int index, T item)
        {
            data.Insert(index, item);
        }

        public void Insert(int index, params T[] items)
        {
            data.InsertRange(index, items);
        }

        public void Insert(int index, IEnumerable<T> items)
        {
            data.InsertRange(index, items);
        }

        public bool Remove(T item)
        {
            var index = Count - 1;
            if (Count > 0 && data[index].Equals(item))
            {
                data.RemoveAt(index);
                return true;
            }
            else
            {
                return false;
            }
        }

        public void RemoveTop()
        {
            data.RemoveAt(Count - 1);
        }

        public void Remove(int count)
        {
            for (var i = 0; i < count; i++)
            {
                data.RemoveAt(Count - 1);
            }
        }

        public void RemoveAt(int index)
        {
            data.RemoveAt(index);
        }

        public void RemoveAt(int index, int count)
        {
            data.RemoveRange(index, count);
        }

        public bool TryPop(out T item)
        {
            if (Count > 0)
            {
                item = Pop();
                return true;
            }
            else
            {
                item = default(T);
                return false;
            }
        }

        IEnumerator IEnumerable.GetEnumerator()
        {
            return (data as IEnumerable).GetEnumerator();
        }

        public override string ToString()
        {
            return string.Join(Delimiter, data);
        }
    }
}
